require 'sinatra/base'
require 'slim'
require 'bunny'
require 'multi_json'
require 'oj'
require 'url_encryptor'
require 'typhoeus'
require 'digest'

class ActionServer < Sinatra::Base
  QUEUE_NAME = 'action'

  class << self
    attr_reader :rabbitmq_config, :queue, :cache_dir

    def rabbitmq_config=(config)
      @rabbitmq_config = config
    end

    def cache_dir=(path)
      @cache_dir = path
    end

    def channel
      @channel ||= begin
        connection = Bunny.new(
          host:     rabbitmq_config[:host],
          user:     rabbitmq_config[:user],
          password: rabbitmq_config[:password],
          vhost:    rabbitmq_config[:vhost]
        )
        connection.start
        connection.create_channel
      end
    end

    def exchange
      @exchange ||= channel.default_exchange
    end
  end

  def cache_dir
    self.class.cache_dir
  end


  def publish(message)
    self.class.exchange.publish(message, routing_key: QUEUE_NAME)
  end

  def extract_ip(request)
    ip = request['HTTP_X_FORWARDED_FOR'] || request.ip
    ip.to_s
  end

  def extract_headers(request)
    headers = {}

    request.env.each do |name, val|
      header = name.match(/^HTTP_(.*)$/)
      if header && header[1]
        headers[header[1]] = val
      end
    end

    headers
  end

  def md5(string)
    md5 = Digest::MD5.new
    md5.update(string)
    md5.hexdigest
  end

  def downloaded_file(src, dest)
    downloaded_file = nil
    request = Typhoeus::Request.new(src, followlocation: true)

    request.on_headers do |response|
      if response.code == 200
        downloaded_file = File.open(dest, 'wb')
      else
        error response.code
      end
    end

    request.on_body do |chunk|
      downloaded_file.write(chunk) if downloaded_file
    end

    request.on_complete do |response|
      downloaded_file.close if downloaded_file
    end

    request.run
  end

  set :root,  File.expand_path(File.join(File.dirname(__FILE__), '../'))
  set :views, File.join(settings.root, 'views')
  set :public_folder, File.join(settings.root, 'static')

  not_found do
    status 404
  end

  error OpenSSL::Cipher::CipherError do
    status 404
  end

  get '/click/:data' do
    query = UrlEncryptor.decrypt_query_string(params[:data])

    message = MultiJson.dump(
      action:  'click',
      query:   query,
      headers: extract_headers(request)
    )
    publish(message)

    redirect query["url"]
  end

  get '/open/:data' do
    query = UrlEncryptor.decrypt_query_string(params[:data])

    message = MultiJson.dump(
      action:  'open',
      query:   query,
      headers: extract_headers(request)
    )
    publish(message)

    file_path = File.join(settings.public_folder, 'pixel.png')
    send_file file_path, type: 'image/png', disposition: 'inline', filename: 'pixel.png'
  end

  get '/proxy/:data' do
    query = UrlEncryptor.decrypt_query_string(params[:data])

    file_name = "#{ md5(query["url"]) }.cache"
    file = File.join(cache_dir, file_name)
    unless File.exists?(file)
      downloaded_file(query["url"], file)
    end

    send_file file
  end
end
